// watchdog.c

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <avr/pgmspace.h>
#include <string.h>

uint16_t WdtTrapAddress __attribute__ ((section (".noinit")));      // store watchdog interrupt return address

static uint8_t isrStackOffset;

void initWatchdog(uint8_t timeout) {
    uint8_t sreg = SREG;
    cli();

    if (MCUSR & (1 << PORF))                             // on power on reset
        WdtTrapAddress = 0;
    MCUSR = 0;

    // scan wdt isr for return address offset
    const uint16_t *pCode = (const uint16_t *) (*((const __flash uint16_t *) 0x1a) << 1) + 5;
    uint16_t code;

    // Determine # of stack entries used by watchdog ISR
    #define LOAD(c, p)      c = pgm_read_word(p++)
    isrStackOffset = 4;
    LOAD(code, pCode);
    while (0x920fu == (code & 0xfe0fu)) { // swapped masks 0x0f92u and 0x0ffeu so we do not need to byte swap code
        LOAD(code, pCode);
        isrStackOffset++;
    }
    #undef LOAD

    wdt_enable(timeout);
    WDTCSR |= (1 << WDIE);                              // enable WDT interrupt

    SREG = sreg;
}

void idle(void) {
    uint8_t timeout = (WDTCSR & ((1 << WDP2) | (1 << WDP1) | (1 << WDP0))) | ((WDTCSR & (1 << WDP3) ? 8 : 0));

	wdt_reset();
    WDTCSR |= (1 << WDIE);                              // enable WDT interrupt

    sleep_cpu();

	wdt_reset();
    wdt_enable(timeout);                                // remember timeout
    WDTCSR |= (1 << WDIE);                              // enable WDT interrupt
}

#define GET_ISR_RETADDR()   (((*(uint8_t *) ((uint8_t *) SP + isrStackOffset)) << 8) | (*(uint8_t *) ((uint8_t *) SP + isrStackOffset + 1))) << 1

ISR(WDT_vect) {
    if ((_SLEEP_CONTROL_REG & (uint8_t)_SLEEP_ENABLE_MASK)) {
        // watchdog interrupt in idle
        wdt_disable();                                  // prevent reset
    } else {
        // error condition
        WdtTrapAddress = GET_ISR_RETADDR();             // store ISR return address in .noinit section, so it will overcome the reset

        #ifdef ERROR_ShutdownHook
            ERROR_ShutdownHook;
        #endif // ERROR_ShutdownHook

        // *** RESET MCU NOW ***
        while (1) ;                                     // wait for 2nd watchdog overflow
    }
}
